#!/usr/bin/env python3

# Filename:     test_meshtomeshregistration.py
# Contributors: Tom Birdsong
# Purpose:  Pytest definitions for mesh-to-mesh registration classes
#           defined in the src/hasi/hasi module

# Directions:
# - cd into the src/hasi directory
# - run `pytest .`

import os
import sys
from urllib.request import urlretrieve
module_path = os.path.abspath(os.path.join('.'))

if module_path not in sys.path:
  sys.path.append(module_path)

import itk

MEANSQUARES_METRIC_MAXIMUM_THRESHOLD = 0.002
DIFFEO_METRIC_MAXIMUM_THRESHOLD = 0.0001
MAX_ITERATIONS = 200

FIXED_MESH_FILE = 'test/Input/901-L-mesh.vtk'
MOVING_MESH_FILE = 'test/Input/901-R-mesh.vtk'

os.makedirs('test', exist_ok=True)
os.makedirs('test/Input', exist_ok=True)
os.makedirs('test/Output', exist_ok=True)

if not os.path.exists(FIXED_MESH_FILE):
    url = 'https://data.kitware.com/api/v1/file/5f9daaae50a41e3d1924dae1/download'
    urlretrieve(url, FIXED_MESH_FILE)
    
if not os.path.exists(MOVING_MESH_FILE):
    url = 'https://data.kitware.com/api/v1/file/5f9daaae50a41e3d1924dae1/download'
    urlretrieve(url, MOVING_MESH_FILE)

os.makedirs('test/Output', exist_ok=True)

mesh1 = itk.meshread(FIXED_MESH_FILE)
mesh2 = itk.meshread(MOVING_MESH_FILE)

# Test base class import and functions
def test_mesh_to_image():
    # Class is imported
    from hasi.meshtomeshregistrar import MeshToMeshRegistrar

    # Class is instantiable
    registrar = MeshToMeshRegistrar()

    # Initialize runs without error
    registrar.initialize()

    # Mesh-to-image executes without error
    img1 = registrar.mesh_to_image(mesh1)

    # Image was generated
    assert(type(img1) == itk.Image[itk.F, 3])

    # Mesh-to-image can run with reference image
    img2 = registrar.mesh_to_image(mesh2, img1)

    # Image was generated
    assert(type(img2) == itk.Image[itk.F, 3])

    # Images occupy the same physical space
    assert(img2.GetOrigin() == img1.GetOrigin())
    assert(img2.GetSpacing() == img1.GetSpacing())

# Test registration with mean squares image metric
def test_meansquares_registration():
    MESH_OUTPUT = 'test/Output/testMeanSquaresRegisterOutput.vtk'

    # Class is imported
    from hasi.meansquaresregistrar import MeanSquaresRegistrar

    # Class is instantiable
    registrar = MeanSquaresRegistrar()

    # Registration executes
    mesh3 = registrar.register(mesh1,mesh2,filepath=MESH_OUTPUT)

    # Mesh was generated
    assert(type(mesh3) == type(mesh1))

    # Optimization converged
    assert(registrar.optimizer.GetCurrentMetricValue() <
            MEANSQUARES_METRIC_MAXIMUM_THRESHOLD)
  
    # Optimization did not exceed allowable iterations
    assert(registrar.optimizer.GetCurrentIteration() <= MAX_ITERATIONS)
  
    # Mesh was output
    assert(os.path.isfile(MESH_OUTPUT))

    # File output matches mesh
    fileMesh3 = itk.meshread(MESH_OUTPUT)
    assert(fileMesh3.GetNumberOfPoints() == mesh3.GetNumberOfPoints())
    assert(all(mesh3.GetPoint(i) == fileMesh3.GetPoint(i)
                for i in range(0,mesh3.GetNumberOfPoints())))

    # Clean up
    os.remove(MESH_OUTPUT)

# Test registration with diffeomorphic demons image registration
def test_diffeo_registration():
    MESH_OUTPUT = 'test/Output/testDiffeoRegisterOutput.vtk'

    # Class is imported
    from hasi.diffeoregistrar import DiffeoRegistrar

    # Class is instantiable
    registrar = DiffeoRegistrar()

    # Registration executes without error
    mesh4 = registrar.register(mesh1,mesh2,filepath=MESH_OUTPUT)

    # Mesh was generated
    assert(type(mesh4) == type(mesh1))

    # Mesh was output
    assert(os.path.isfile(MESH_OUTPUT))

    # Optimization converged
    assert(registrar.filter.GetMetric() <
            DIFFEO_METRIC_MAXIMUM_THRESHOLD)

    # Optimization did not exceed allowable iterations
    assert(registrar.filter.GetElapsedIterations() <= MAX_ITERATIONS)

    # Clean up
    os.remove(MESH_OUTPUT)



#!/usr/bin/env python3

# Filename: MeanSquaresRegistrar.py
# Contributors: Tom Birdsong
# Purpose: Python implementation of Slicer SALT mesh-to-mesh registration Cxx
# CLI.
#    Ported from https://github.com/slicersalt/RegistrationBasedCorrespondence/

import logging
import os

import itk

from .meshtomeshregistrar import MeshToMeshRegistrar

class MeanSquaresRegistrar(MeshToMeshRegistrar):
    # Common definitions
    MAX_ITERATIONS = 200

    Dimension = 3
    PixelType = itk.F
    ImageType = itk.Image[PixelType, Dimension]
    TemplateImageType = ImageType
    TargetImageType = ImageType

    MaskPixelType = itk.UC
    MaskImageType = itk.Image[MaskPixelType, Dimension]
    FixedPointSetType = itk.PointSet[itk.F, Dimension]

    MeshType = itk.Mesh[itk.F, Dimension]

    # Definitions for default registration
    V_SPLINE_ORDER = 3
    GRID_NODES_IN_ONE_DIMENSION = 4

    COST_FN_CONVERGENCE_FACTOR = 1e7
    GRADIENT_CONVERGENCE_TOLERANCE = 1e-35
    MAX_FUNCTION_EVALUATIONS = 200
    MAX_CORRECTIONS = 7

    NUMBER_OF_LEVELS = 1
    SMOOTHING_SIGMAS_PER_LEVEL = [0] * NUMBER_OF_LEVELS
    SHRINK_FACTORS_PER_LEVEL = [1] * Dimension

    # TODO investigate if this is useful
    LinearInterpolatorType = \
        itk.LinearInterpolateImageFunction[ImageType, itk.D]
    

    def __init__(self):
        super(self.__class__,self).__init__()

    def initialize(self):
        # Optimizer is exposed so that calling scripts can reference with custom observers
        self.optimizer = itk.LBFGSBOptimizerv4.New(CostFunctionConvergenceFactor=self.COST_FN_CONVERGENCE_FACTOR,
            GradientConvergenceTolerance=self.GRADIENT_CONVERGENCE_TOLERANCE,
            MaximumNumberOfFunctionEvaluations=self.MAX_FUNCTION_EVALUATIONS,
            MaximumNumberOfCorrections=self.MAX_CORRECTIONS)


    # Register two 3D images with an LBFGSB optimizer
    def register(self,
                 template_mesh:MeshType,
                 target_mesh:MeshType,
                 filepath:str=None,
                 verbose=False,
                 num_iterations=MAX_ITERATIONS) -> MeshType:

        template_image = self.mesh_to_image(template_mesh)
        target_image = self.mesh_to_image(target_mesh, template_image)

        ImageType = type(template_image)
        Dimension = template_image.GetImageDimension()
            
        metric = itk.MeanSquaresImageToImageMetricv4[ImageType,ImageType].New()

        # Set physical dimensions for transform
        fixed_physical_dimensions = list()
        fixed_origin = list(template_image.GetOrigin())

        for i in range(0, Dimension):
            fixed_physical_dimensions.append(template_image.GetSpacing()[i] * \
                (template_image.GetLargestPossibleRegion().GetSize()[i] - 1))
            
        TransformType = itk.BSplineTransform[itk.D, Dimension, self.V_SPLINE_ORDER]
        transform = TransformType.New(TransformDomainOrigin=fixed_origin,
                                            TransformDomainPhysicalDimensions=fixed_physical_dimensions,
                                            TransformDomainDirection=template_image.GetDirection())
        transform.SetTransformDomainMeshSize([self.GRID_NODES_IN_ONE_DIMENSION - self.V_SPLINE_ORDER] * Dimension)

        # Transform parameters unbounded for typical registration problem
        number_of_parameters = transform.GetNumberOfParameters()
        self.optimizer.SetBoundSelection([0] * number_of_parameters)
        self.optimizer.SetUpperBound([0] * number_of_parameters)
        self.optimizer.SetLowerBound([0] * number_of_parameters)

        # Monitor optimization via observer
        def print_iteration():
            iteration = self.optimizer.GetCurrentIteration()
            metric = self.optimizer.GetCurrentMetricValue()
            infnorm = self.optimizer.GetInfinityNormOfProjectedGradient()
            print(f'{iteration} {metric} {infnorm}')

        if(verbose):
            self.optimizer.AddObserver(itk.IterationEvent(),
                                            print_iteration)

        self.optimizer.SetNumberOfIterations(num_iterations)

        # Define object to handle image registration
        RegistrationType = \
            itk.ImageRegistrationMethodv4[ImageType,ImageType]
        registration = RegistrationType.New(InitialTransform=transform,
            FixedImage=template_image,
            MovingImage=target_image,
            Metric=metric,
            Optimizer=self.optimizer,
            NumberOfLevels=self.NUMBER_OF_LEVELS,
            ShrinkFactorsPerLevel=self.SHRINK_FACTORS_PER_LEVEL)
        registration.SetSmoothingSigmasPerLevel(self.SMOOTHING_SIGMAS_PER_LEVEL)
        registration.InPlaceOn()

        # Run registration
        # NOTE ignore warning: "LBFGSBOptimizer does not support
        #     scaling, all scales set to one"
        #     Registration likely attempts to set scales by default,
        #     no observed impact on performance from this warning
        registration.Update()

        # TODO functional interface
        # image_registration_method()

        # Report results
        if(verbose):
            print('Solution = ' + str(list(transform.GetParameters())))

        # Update fixed image
        # TODO try inverting moving transform and compare results

        transformed_mesh = itk.transform_mesh_filter(template_mesh, transform=transform)

        # Write out
        if filepath is not None:
            itk.meshwrite(transformed_mesh, filepath)
            if(verbose):
                print(f'Wrote resulting mesh to {filepath}')

        return transformed_mesh



#!/usr/bin/env python3

# Filename: MeshToMeshRegistration
# Contributors: Tom Birdsong
# Purpose: Python implementation of Slicer SALT mesh-to-mesh registration Cxx
# CLI.
#    Ported from https://github.com/slicersalt/RegistrationBasedCorrespondence/
import logging
import os

import itk


class MeshToMeshRegistrar:
    # Common definitions
    MAX_ITERATIONS = 200

    Dimension = 3
    PixelType = itk.F
    ImageType = itk.Image[PixelType, Dimension]
    FixedImageType = ImageType
    MovingImageType = ImageType

    MaskPixelType = itk.UC
    MaskImageType = itk.Image[MaskPixelType, Dimension]
    FixedPointSetType = itk.PointSet[itk.F, Dimension]

    MeshType = itk.Mesh[itk.F, Dimension]

    def __init__(self):
        self.initialize()

    # Abstract method for developer to explicitly reset internal state objects
    def initialize(self):
        pass
    
    def register(self, template_mesh:MeshType, target_mesh:MeshType,
                       filepath:str=None, verbose=False) -> MeshType:
        raise NotImplementedError('Base class does not implement register functionality')

    # Convert itk.Mesh object to 3-dimensional itk.Image object
    # Defined for mesh of type itk.F
    def mesh_to_image(self, mesh:MeshType, reference_image:ImageType=None) -> ImageType :

        MeshToImageFilterType = itk.TriangleMeshToBinaryImageFilter[self.MeshType,self.ImageType]

        if not reference_image:
            bounds = mesh.GetBoundingBox().GetBounds()
            spacing = ((bounds[1] - bounds[0]) / 90,
                        (bounds[3] - bounds[2]) / 90,
                        (bounds[5] - bounds[4]) / 90)
            origin = (bounds[0] - 5 * spacing[0],
                        bounds[2] - 5 * spacing[1],
                        bounds[4] - 5 * spacing[2])
            size = (100,100,100)
        
            filter = MeshToImageFilterType.New(Input=mesh,
                                                Origin=origin,
                                                Spacing=spacing,
                                                Size=size)
        else:
            filter = MeshToImageFilterType.New(Input=mesh,
                                                Origin=reference_image.GetOrigin(),
                                                Spacing=reference_image.GetSpacing(),
                                                Size=reference_image.GetLargestPossibleRegion().GetSize())
        filter.Update()

        DistanceType = itk.SignedMaurerDistanceMapImageFilter[self.ImageType,self.ImageType]
        distance = DistanceType.New(Input=filter.GetOutput())
        distance.Update()

        image = distance.GetOutput()
        return image







#!/usr/bin/env python3

# Filename: DiffeoRegistrar.py
# Contributors: Tom Birdsong
# Purpose: Python implementation of Slicer SALT mesh-to-mesh registration Cxx
# CLI.
#    Ported from https://github.com/slicersalt/RegistrationBasedCorrespondence/

import logging
import os

import itk

from .meshtomeshregistrar import MeshToMeshRegistrar

class DiffeoRegistrar(MeshToMeshRegistrar):
    
    # Common definitions
    MAX_ITERATIONS = 200

    Dimension = 3
    PixelType = itk.F
    ImageType = itk.Image[PixelType, Dimension]
    TemplateImageType = ImageType
    TargetImageType = ImageType

    MaskPixelType = itk.UC
    MaskImageType = itk.Image[MaskPixelType, Dimension]
    FixedPointSetType = itk.PointSet[itk.F, Dimension]

    MeshType = itk.Mesh[itk.F, Dimension]

    # Definitions for diffeometric registration
    STANDARD_DEVIATIONS = 1.0


    def __init__(self):
        self.initialize()

    # Expose method to reset persistent objects as desired
    def initialize(self):
        VectorPixelType = itk.Vector[itk.F, self.Dimension]
        DisplacementFieldType = itk.Image[VectorPixelType, self.Dimension]
        DiffeoRegistrationFilterType = \
            itk.DiffeomorphicDemonsRegistrationFilter[self.ImageType,
                                                      self.ImageType,
                                                      DisplacementFieldType]
        self.filter = DiffeoRegistrationFilterType.New(
            StandardDeviations=self.STANDARD_DEVIATIONS)

    # Register meshes with diffeomorphic demons algorithm
    def register(self,
                 template_mesh:MeshType,
                 target_mesh:MeshType,
                 filepath:str=None,
                 verbose=False,
                 max_iterations=MAX_ITERATIONS) -> MeshType:
        template_image = self.mesh_to_image(template_mesh)
        target_image = self.mesh_to_image(target_mesh,template_image)

        self.filter.SetFixedImage(template_image)
        self.filter.SetMovingImage(target_image)

        self.filter.SetNumberOfIterations(max_iterations)
      
        def print_iteration():
            metric = self.filter.GetMetric()
            print(f'{metric}')

        if(verbose):
            self.filter.AddObserver(itk.ProgressEvent(),print_iteration)

        # Run registration
        self.filter.Update()
        
        # Update template mesh to match target
        DisplacementFieldTransformType = \
            itk.DisplacementFieldTransform[itk.F, self.Dimension]
        transform = DisplacementFieldTransformType.New()
        transform.SetDisplacementField(self.filter.GetOutput())

        # TODO pythonic style
        TransformFilterType = \
            itk.TransformMeshFilter[self.MeshType,
                                    self.MeshType,
                                    itk.Transform[itk.F,self.Dimension,self.Dimension]]
        transform_filter = TransformFilterType.New(Input=template_mesh,
            Transform=transform)
        transform_filter.Update()

        if filepath is not None:
            itk.meshwrite(transform_filter.GetOutput(),filepath)
            if(verbose):
                print(f'Wrote resulting mesh to {filepath}')

        return transform_filter.GetOutput()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hasi']

package_data = \
{'': ['*']}

install_requires = \
['itk-hasi']

extras_require = \
{'test': ['pytest >=6.2.1']}

setup(name='hasi',
      version='0.1',
      description='High-throughput Applications for Skeletal Imaging (HASI) Python content',
      author='Kitware Medical',
      author_email='matt.mccormick@kitware.com',
      url='https://github.com/KitwareMedical/HASI',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
